#!/usr/bin/env ruby
require 'rubygems'
require 'rmagick'
require 'pp'

###
# Converts a gif file to c code to enable displaying animations.
#
# The data format is described in animations.h
###

def compress(data, use_cutoff=true)
	cutoff = 3
	escape = 255
	last = data[0]
	result = []
	count = 0
	data.each do |x|
		if x == last
			count += 1
		end
		
		if x!=last || count==255
			if use_cutoff
				if count <= cutoff
					count.times { result << last }
				else
					result << escape << count << last
				end
			else
				result << count << last
			end
			count = 1
			last = x
		end
	end
	if use_cutoff
		if count < cutoff
			count.times { result << last }
		else
			result << escape << count << last
		end
	else
		result << count << last
	end
	return result
end
		
unless ARGV[0] && ARGV[1]
	puts "Usage:"
	puts "#{$0} <gif_file> <variable_name>"
	puts "e.g. '#{$0} mario.gif mario'"
	puts
	puts "Resulting data format is described in animations.h"
	exit 1
end	

image_file=ARGV[0]
name=ARGV[1]
frames = Magick::ImageList.new(image_file)

STDERR.puts "Found #{frames.count} frames."

STDERR.puts "Getting delays..."
ticks_per_second = frames.ticks_per_second
times = [] ; frames.each{|f| times << (1000.0 / ticks_per_second * f.delay).round}
individual_frame_times = true
if times.uniq.count==1
	individual_frame_times = false
	times = times[0, 1]
end

STDERR.print "Getting colors..."
colors = []
frames.each do |frame|
	frame.columns.times do |x|
		frame.rows.times do |y|
			color = frame.pixel_color(x, y).to_color(Magick::AllCompliance, true, 8, true)[1,8].to_i(16)
			colors << color
		end
	end
end
colors = colors.uniq

STDERR.puts " Found #{colors.count} colors."

transparent = colors.select{|c| c & 0xFF == 0}

STDERR.puts "#{transparent.count} color(s) being transparent."

# color[0] is "keep the color from the previous frame"
# color[1] is "background color"
colors = ([0x00000012, 0x00000013] + (colors - transparent)).uniq
STDERR.puts "Using #{colors.count} colors."
raise "Number of colors has to be 255 or less!" if colors.count>255

STDERR.puts
puts
puts "uint8_t animation_#{name}_colors[] PROGMEM = {#{colors[2..-1].map{|c| [c>>24 & 0xFF, c>>16 & 0xFF, c>>8 & 0xFF]}.flatten.join(", ")}};"

p_frame = nil
frames_data = []
times = []
frames.each_with_index do |frame, index|
	data = []
	if index==0 # first frame
		frame.rows.times do |y|
			frame.columns.times do |x|
				color = frame.pixel_color(x, y).to_color(Magick::AllCompliance, true, 8, true)[1,8].to_i(16)
				if transparent.include? color
					data << 1
				else
					data << colors.index(color)
				end
			end
		end
	else
		frame.rows.times do |y|
			frame.columns.times do |x|
				color = frame.pixel_color(x, y).to_color(Magick::AllCompliance, true, 8, true)[1,8].to_i(16)
				p_color = p_frame.pixel_color(x, y).to_color(Magick::AllCompliance, true, 8, true)[1,8].to_i(16)
				
				if color==p_color
					data << 0
				elsif transparent.include? color
					data << 1
				else
					data << colors.index(color)
				end
			end
		end
	end
	time = (1000.0 / ticks_per_second * frame.delay).round
	if frame==0 || data.uniq!=[0]
		times << time
		frames_data << data
	else
		old_time = times.pop
		times << old_time + time
	end
	p_frame = frame
end

data = frames_data.map{|d| compress(d, true)}

puts "uint8_t animation_#{name}_data[] PROGMEM = {\n  #{data.map{|d| d.join(",")}.join(",\n  ")}\n};"
puts "uint16_t animation_#{name}_delays[] = {#{times.join(",")}};"
s=0
puts "uint16_t animation_#{name}_offsets[] = {#{(data.map{|d| t=s; s+=d.count; t} + [s]).join(",")}};"

puts "AnimationData animation_#{name} = {&animation_#{name}_colors[0], &animation_#{name}_data[0], &animation_#{name}_offsets[0], &animation_#{name}_delays[0], #{individual_frame_times}, #{colors.count-2}, #{frames_data.count}, #{frames.first.columns}, #{frames.first.rows}};"
puts
STDERR.puts
STDERR.puts "Space usage:"
STDERR.puts "  Colors:      %6d bytes." % [s1=(colors.count-2) * 3] # colors are 3-bytes, but we have to use uint32_t, which takes up 4 bytes.
STDERR.puts "  Data:        %6d bytes." % [s2=data.flatten.count]
STDERR.puts "  Delays:      %6d bytes." % [s5=times.count * 3 + 1]
STDERR.puts "  Offsets:     %6d bytes." % [s3=data.count * 2]
STDERR.puts "  TOTAL:       %6d bytes." % [s1+s2+s3+s5]
STDERR.puts "Original size: %6d bytes." % [s4=File.new(image_file).size]
STDERR.puts "Difference:    %6d bytes." % [s1+s2+s3+s5 - s4]
